#!/usr/bin/perl
#
#
# Usage:
#
# query.pl --game q3|d3 --command command [--rcon] [--loop number] [--delay number] [--debug] host port
#
#
# Options:
# 
# --game     specify q3 for quake3 or d3 for doom3
# --command  depends on wether you use rcon or not; examples:
#            "heartbeat" (d3 and rcon)
#            "say Server shutting down." (d3|q3 and rcon)
#            getInfo (d3 without rcon)
#            getstatus (q3 without rcon)
#            With rcon basically every server command is valid
# --rcon     send rcon command (you'll be asked for the password)
# --loop     how often to send the command; 0 means indefinitely
# --delay    sleep time between loop iterations
# --debug    some extra output
#
#
# Examples:
#
# query.pl --game d3 --command "heartbeat" --rcon --loop 0 --delay 60 somehost someport
#  Force d3 server to do a heartbeat every minute.
# query.pl --game q3 --command "status" --rcon --loop 1 somehost someport
#  Get q3 server status once.
# query.pl --game q3 --command "getstatus" --loop 1 somehost someport
#  Get 'raw' q3 server status once.
#
#
# Notes:
#
# Please note that the packet payloads will not be decoded. This means, while
# rcon response payloads are generally in readable form, non-rcon responses are
# not. E.g. when you send a non-rcon getinfo to a d3 server, the response is
# roughly readable, but it contains nonprintable chars and needs further decoding
# to be actually readable.
#
# Since this script is meant to be a general framework to send any kind of query
# to q3/d3 servers, I consider response decoding a special requirement, so please
# do it yourself in case of need.
#
# Regarding the rcon password, it is read from stdin with no echo. I didn't provide
# a command line option since it is generally a bad idea to provide passwords on the
# command line. If you want to prevent interaction, you may pipe the password
# via echo or cat into the script, like
# cat some-file-containing-passwd | query.pl ...
#
#
# Author : Bruno Hertz (brrhtz at yahoo dot de)
# License: GPL
#
#
# Modifications:
#
# Inititial release 22 Oct 2004
#
#
# Todo:
#
# Maybe document the package format :)
#
#


use Getopt::Long;
use Term::ReadKey;
use IO::Socket::INET;
use IO::Select;

my $game="";             # game: q3 or d3
my $command;             # query command: status, getinfo etc.
my $rcon;                # flag wether this is an rcon command
my $loop=1;              # loop count - 0 means indefinitely
my $loop_delay=60;       # sleep time for each loop
my $debug;               # extra output

my $host;                # host to query
my $port;                # query port

my $passwd;              # rcon password, used only for rcon commands
my $timeout=2;           # socket read timeout
my $loop_indefinitely=0; # loop forever?

my @prefix;              # UDP packet prefix (couple of 0xFF)
my $sep;                 # Send packet field separator
my $print_suffix;        # Suffix spec for suffix sent by server after print cmd.
my $packstring;          # Pack template for UDP send packet
my $unpackstring;        # Unpack template for UDP send packet
my $query;               # Packed string which will actually be sent over network
my @query;               # Array of fields in query string, for debugging only
my $response;            # UDP read buffer
my @response;            # Array of fields in response string

my $socket;              # Socket handle
my $select;              # Select handle

# Process command line.
GetOptions ('game=s'    => \$game,
            'command=s' => \$command,
            'loop=i'    => \$loop,
            'delay=i'   => \$loop_delay,
            'rcon'      => \$rcon,
            'debug'     => \$debug)
    or die "Error parsing options.";

($host, $port) = @ARGV;

# Checks.
$host or die "No host specified.\n";
$port or die "No port specified.\n";
$game or die "No game specified.\n";
$command or die "No command specified.\n";

# Loop indefinitely if loop==0.
$loop==0 and $loop=1 and $loop_indefinitely=1;

# Specify UDP packet structures according to gametype.
if($game eq "q3") {

    @prefix=(255,255,255,255);
    $sep=32;
    $print_suffix="C1";

}
elsif ($game eq "d3") {

    @prefix=(255,255);
    $sep=0;
    $print_suffix="C5";

}
else {die "Unknown game: $game\n"}

# Read rcon passwd from stdin if necessary.
if($rcon) {

    print "Password: ";
    ReadMode 'noecho';
    $passwd = ReadLine 0;
    chomp $passwd;
    ReadMode 'normal';
    $passwd or die "No rcon password specified.\n";
    print "\n";

}

# Pack UDP packet to send (and an unpack to reread it for debugging).
if($rcon) {
    $packstring="C" . scalar(@prefix) . "A4CA". length($passwd) . "CA" . length($command) . "C";
    $unpackstring="C". scalar(@prefix) . "A5" . $print_suffix . "A*";
    $query = pack($packstring,@prefix,"rcon",$sep,$passwd,$sep,$command,0);
}
else {
    $packstring="C" . scalar(@prefix) . "A" . length($command) . "C";
    $unpackstring="C". scalar(@prefix) . "A*";
    $query = pack($packstring,@prefix,$command,0);
}

# Print what we got.
if($debug) {

    print "Host     : $host\n";
    print "Port     : $port\n";
    print "Game     : $game\n";
    print "Command  : $command\n";
    print "Rcon     : $rcon\n";
    print "Loop     : $loop\n";
    print "Delay    : $loop_delay\n";
    print "Password : $passwd\n";
    @query = unpack($packstring, $query );
    print "Sending  : @query\n\n";
}

# UDP socket to server.
$socket = IO::Socket::INET->new(PeerAddr => $host, PeerPort => $port, Proto    => "udp")
          or die "Can't open UDP socket to $host on port $port\n";

# We use select to control the read timeout.
$select = IO::Select->new();
$select->add($socket);

# Main loop.
do {

    $socket->send($query);
    while($select->can_read($timeout))
    {
        $socket->recv($response,4096);
        if($response ne '')
        {
            # Unpack response string and print actual payload.
            @response=unpack($unpackstring,$response);
            print pop(@response) . "\n";
        }
    }
    $loop_indefinitely or $loop--;
    $loop and sleep $loop_delay;

} while $loop;

$socket->close;
